// {{{ preprocessors
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdint.h>
#include <libdragon.h>
#include <math.h>
#include <array>
#include <vector>
#include <mikmod.h>

#include "../../include/regsinternal.h"

#include "drv_n64.h"
#include "sram.h"

// TODO to be deleted! {{{
#ifdef __cplusplus
extern "C" {
#endif
#include <PR/mbi.h> // M_???TASK, GBI(, ABI)
#include <PR/sptask.h> // OSTask
#include <PR/ucode.h> // ucode text/data symbols
#include <PR/gu.h> // gu*
#ifdef __cplusplus
}
#endif

#ifndef GPACK_RGBA8888
#define GPACK_RGBA8888(r, g, b, a) ((((r) & 0xFF) << 24) | (((g) & 0xFF) << 16) | (((b) & 0xFF) << 8) | (((a) & 0xFF) << 0))
#endif
// }}}

#include "common.h"
#include "rsp.h"
#include "polydata.h"
#include "mtx.h"

#ifndef countof
#define countof(a) (sizeof(a)/sizeof((a)[0]))
#endif

extern "C" void *sbrk(int incr); // at system.c but not exported
extern "C" void _exit(int);
// }}}

// {{{ math
inline float deg2rad(float deg) {
	return deg / 180.0f * M_PI;
}
inline float rad2deg(float rad) {
	return rad / M_PI * 180.0f;
}

/* note: current and targets must be normalized. */
float nearside(float current, float target) {
	if(180.0f < current) {
		if(target < current - 180.0f) {
			//target += 360.0f;
			return 1.0f;
		}
	} else {
		if(current + 180.0f < target) {
			//target -= 360.0f;
			return -1.0f;
		}
	}
	if(current < target) {
		return 1.0f;
	} else {
		return -1.0f;
	}
}

// http://d.hatena.ne.jp/gikan/20080507/p1
float degnormalize(float deg) {
	if(0.0f <= deg && deg < 360.0f) {
		return deg;
	}
	deg -= static_cast<int>(deg / 360.0f) * 360.0f;
	if(deg < 0.0f) { // -360 < deg < 0
		deg += 360.0f;
	}
	return deg;
}

float smallerdeg(float current, float target) {
	if(180.0f < current) {
		if(target < current - 180.0f) {
			//return target + 360.0f - current;
			target += 360.0f;
		}
	} else {
		if(current + 180.0f < target) {
			//return current - (target - 360.0f);
			target -= 360.0f;
		}
	}
	if(current < target) {
		return target - current;
	} else {
		return current - target;
	}
}
// }}}

// {{{ definitions
struct Point {
	float x, y;

	Point() {}
	Point(float x, float y) : x(x), y(y) {}
	Point &operator +=(const Point &other) {{{
		x += other.x;
		y += other.y;
		return *this;
	}}}
	Point &operator -=(const Point &other) {{{
		x -= other.x;
		y -= other.y;
		return *this;
	}}}
	float distance(const Point &other) const {{{
		float dx = other.x - x;
		float dy = other.y - y;
		return sqrt((dx*dx) + (dy*dy));
	}}}
	bool incircle(const Point &other, float radius) const {{{
		return distance(other) < radius;
	}}}
	void add_angular(float rotr, float len) {{{
		x += cos(rotr) * len;
		y += sin(rotr) * len;
	}}}
};

struct Rect {
	Point ul, br;

	Rect() {}
	Rect(float ulx, float uly, float brx, float bry) : ul(ulx, uly), br(brx, bry) {}
	bool inrect(const Point &other) const {{{
		return ul.x <= other.x && ul.y <= other.y && other.x < br.x && other.y < br.y;
	}}}
};

class SND1 {
	SAMPLE *sample;
	SBYTE ch;

	const static SBYTE STOPPED = -1, ONESHOT = -2;

public:
#ifdef DEBUG
	static int debugsndcount;
#endif

	void load(SAMPLE *sample, bool oneshot=false) {{{
		this->sample = sample;
		ch = oneshot ? ONESHOT : STOPPED;
	}}}
	void play(int freqdelta=0) {{{
		// note: assuming sample is Transp=G-3 Finetune=-28.
		// TODO: use INSTRUMENT*->samplenote,SAMPLE*->speed
		if(ch != STOPPED && ch != ONESHOT) {
			return;
		}
#if 0
		Voice_Play(8/*ch*/, g_slaser, 0);
		Voice_SetFrequency(8, 553);
		Voice_SetVolume(8, 128);
		Voice_SetPanning(8, PAN_CENTER);
#else
		SBYTE cha = Sample_Play(sample, 0, SFX_CRITICAL);
		Voice_SetFrequency(cha, 22046 + freqdelta);
		if(ch != ONESHOT) {
			ch = cha;
		}
#endif
#ifdef DEBUG
		debugsndcount++;
#endif
	}}}
	void stop() {{{
		if(ch == STOPPED) {
			return;
		}
		Voice_Stop(ch);
		ch = STOPPED;
#ifdef DEBUG
		debugsndcount--;
#endif
	}}}
	void refresh() {{{
		if(ch == STOPPED) {
			return;
		}
		if(Voice_Stopped(ch)) {
			ch = STOPPED;
		}
	}}}
};

enum BulletType {
	BT_PLAYER,
	BT_WEAK,
	BT_STRONG,
	BT_SUPERSTRONG,
	BT_CHASE,
	BT_PLAYERHIT,
	BT_ENEMYHIT,
};

struct Bullet {
	bool enabled;
	Point pt;
	float rot;
	Point delta; // cache
	enum BulletType type;
	int count; // general purpose counter

	Bullet() : enabled(false) {}
	void enable(BulletType type, Point pt, float rot, float speed) {{{
		enabled = true;
		this->type = type;
		this->pt = pt;
		this->rot = rot;
		this->delta.x = speed * cos(deg2rad(rot));
		this->delta.y = speed * sin(deg2rad(rot));
	}}}
};

enum TankType {
	TT_PLAYER,
	TT_WEAK,
	TT_STRONG,
	TT_CHASE,
	TT_CHASE_FAST,
	TT_SUPER,
};

enum BoomState {
	BS_NONE,
	BS_EXPLODING, // chain
	BS_IMPLODING, // no chain
	BS_DONE,
};

struct Tank {
	bool enabled;
	Point pt;
	enum TankType type;
	int count[2]; // general purpose counter
	enum BoomState boomstate;
	int boomviewcount, boomstatecount;
	float rot, headrot;
	int chain;

	Tank() : enabled(false) {}
	void enable(TankType type, Point pt, float rot) {{{
		enabled = true;
		this->type = type;
		this->pt = pt;
		this->rot = this->headrot = rot;
		this->boomstate = BS_NONE;
		this->chain = 1;
	}}}
	void boom() {{{
		// TODO assert boomstate == BS_NONE
		if(boomstate != BS_NONE) {
			return;
		}
		boomstate = BS_EXPLODING;
		boomstatecount = 0;
		boomviewcount = 0;
	}}}
};

struct Block {
	Point pt;
	Rect rc; // cache for hittest

	Block(Point pt) {
		this->pt = pt;
		// TODO cache rc
	}
};

class Scene {
public:
	virtual void init() = 0;
	virtual void input() = 0;
	virtual void think() = 0;
	virtual void draw(Gfx **p) = 0;
};

class GameScene : public Scene {
#ifdef DEBUG
public:
#endif
	std::array<Tank, 1+8> tanks;
	std::array<Bullet, 8+7*4> bullets;
	int entryblkz;
	size_t mybullets, mybullet_color;
	std::vector<Block> blocks;
	int gametimer, fintimer, reftimer, engtimer;
	bool guarding, shooting;
	bool moving;
	float myspeed;
	float mymomentumrot;

public:
	// constexpr? if const,
	// error: 'constexpr' needed for in-class initialization of static data member 'const float GameScene::XXX' of non-integral type [-fpermissive]
	constexpr static float GE_FULL = 100.0f, GE_WARNLINE = 50.0f, GE_ALERT = 23.0f, GE_REDLINE = 20.0f;
	constexpr static float GE_PENALTY = 2.0f;
	constexpr static float GED_IDLEREGEN = 0.3f;
	constexpr static float GED_SHOOTING = 0.1f, GED_GUARDING = 0.1f, GED_LASER = 4.0f, GED_BOOM = 4.0f;
	constexpr static float GEM_MOVING = 3;

	virtual void init();
	virtual void input();
	virtual void think();
	virtual void draw(Gfx **p);

protected:
	bool canmove(Tank *me, Point target, bool *pslow);
	// FIXME ugly "doslow", it should be handled by think_weak or something. "myspeed" for enemies?
	bool can_move_then_move(Tank *me, float rotr, float speed, bool *pslow, float olen, float slowcoeff);
	void bullethit(Bullet *bullet, bool quick=false);
	bool isfin();
	void fincheck();
	void boom(Tank *tank, int chain=0);
	void hittest(Bullet *bullet);
	void think_weak(Tank *tank);
};

class PrebattlingScene : public Scene {
	int count;

public:
	virtual void init();
	virtual void input();
	virtual void think();
	virtual void draw(Gfx **p);
};

class GameoverScene : public Scene {
	int count;

public:
	virtual void init();
	virtual void input();
	virtual void think();
	virtual void draw(Gfx **p);
};

class TitleScene : public Scene {
	int bcount;
	int flipcount;
	int count;
	int reset_timer;

public:
	virtual void init();
	virtual void input();
	virtual void think();
	virtual void draw(Gfx **p);
};
// }}}

// {{{ globals
static int vsync = 0;

static void vsync_handler() {
	vsync = 1;
}

extern const char _binary_Module1_xm_start[], _binary_Module1_xm_size[]; // defined by objcopy.
extern const char _binary_fontia44_raw_start[];
extern const char _binary_player_data_start[];

static SND1 g_slaser, g_senedown, g_sgetready, g_smyguard, g_salert, g_smyref1, g_smyref2, g_sextend, g_shighscore, g_sengine;

static int g_score, g_highscore;
static int g_lives;
static int g_count;
static int g_battling;
static float g_energy;
static Scene *g_curscene, *g_nextscene;
static bool g_highscored;

static GameScene g_scene_game;
static PrebattlingScene g_scene_prebattling;
static GameoverScene g_scene_gameover;
static TitleScene g_scene_title;

static int eyex, eyey, eyez; // should be removed?
#ifdef DEBUG
static bool debugon = true;
static char pdebugx, pdebugy;
static int debuggfxcount, debuggfxmax;
int SND1::debugsndcount;
#endif

static const int SC_FIRSTEXT = 15000, SC_EVERYEXT = 15000;

static const Mtx g_mtx_e = { { {0x00010000, 0x00000000, 0x00000001, 0x00000000}, {0x00000000, 0x00010000, 0x00000000, 0x00000001}, {0,0,0,0}, {0,0,0,0} } };
static Mtx g_mtx_move_right, g_mtx_move_down;
static Vtx g_vs_rect16[4] = {
	{ { {  0,  0, 0 }, 0, {         0,         0 }, { 0, 0, 0, 255 } } },
	{ { { 16,  0, 0 }, 0, { (16*2)<<5,         0 }, { 0, 0, 0, 255 } } },
	{ { { 16, 16, 0 }, 0, { (16*2)<<5, (16*2)<<5 }, { 0, 0, 0, 255 } } },
	{ { {  0, 16, 0 }, 0, {         0, (16*2)<<5 }, { 0, 0, 0, 255 } } },
};
static Vtx g_vs_rect32[4] = {
	{ { {  0,  0, 0 }, 0, {         0,         0 }, { 0, 0, 0, 255 } } },
	{ { { 16,  0, 0 }, 0, { (32*2)<<5,         0 }, { 0, 0, 0, 255 } } },
	{ { { 16, 16, 0 }, 0, { (32*2)<<5, (32*2)<<5 }, { 0, 0, 0, 255 } } },
	{ { {  0, 16, 0 }, 0, {         0, (32*2)<<5 }, { 0, 0, 0, 255 } } },
};
static Vtx g_vs_rect1[4] = {
	{ { { 0, 0, 0 }, 0, { 0, 0 }, { 0, 0, 0, 255 } } },
	{ { { 1, 0, 0 }, 0, { 0, 0 }, { 0, 0, 0, 255 } } },
	{ { { 1, 1, 0 }, 0, { 0, 0 }, { 0, 0, 0, 255 } } },
	{ { { 0, 1, 0 }, 0, { 0, 0 }, { 0, 0, 0, 255 } } },
};

struct Color {
	u8 r, g, b;
};

static const Color g_rainbowcolors[8] = {
	{ 0xff, 0x00, 0x00 },
	{ 0xff, 0x80, 0x00 },
	{ 0xff, 0xff, 0x00 },
	{ 0x00, 0xff, 0x00 },
	{ 0x00, 0xff, 0xff },
	{ 0x60, 0x60, 0xff },
	{ 0x00, 0x00, 0xff },
	{ 0xff, 0x00, 0xff },
};

MODULE *g_module;

OSTask gfxtask;
u64 dram_stack[SP_DRAM_STACK_SIZE64] __attribute__((aligned(16)));
u64 output_buf[0x1000 / sizeof(u64)] __attribute__((aligned(16)));
Gfx glist[8192] __attribute__((aligned(16)));

static union {
	uint64_t dw[0x80 / sizeof(uint64_t)];
	struct {
		int highscore;
	} s;
} g_save;
char assert_a[0x80 - sizeof(g_save)]; /* if union "g_save" exceeds max size, complie error happens at here. */

/*
static u32 my_ucode_boot[] __attribute__((aligned(16))) = { 0x0000000B }; // break
static u32 my_ucode[SP_UCODE_SIZE / sizeof(u32)] __attribute__((aligned(16))) = { };
static u32 my_ucode_data[SP_UCODE_DATA_SIZE / sizeof(u32)] __attribute__((aligned(16))) = { };
*/

static short zbuffer[320 * 240];
extern void *__safe_buffer[];

#ifdef DEBUG
extern Mtx g_mtxbuf[256], *g_curmtx;
#endif
// }}}

void scene_next(Scene *scene) {{{
	g_nextscene = scene;
}}}

void draw_init() {{{
	guTranslate(&g_mtx_move_right, 10.0, 0, 0);
	data_cache_hit_writeback(&g_mtx_move_right, sizeof g_mtx_move_right);
	guTranslate(&g_mtx_move_down, 0, 16.0, 0);
	data_cache_hit_writeback(&g_mtx_move_down, sizeof g_mtx_move_down);
}}}

void draw_setup(Gfx **p, int width, int height) {{{
	// rdp init
	gDPPipelineMode((*p)++, G_PM_NPRIMITIVE);
	gDPSetScissor((*p)++, G_SC_NON_INTERLACE, 0, 0, width, height);
	gDPSetTextureLOD((*p)++, G_TL_TILE);
	gDPSetTextureLUT((*p)++, G_TT_NONE);
	gDPSetTextureDetail((*p)++, G_TD_CLAMP);
	gDPSetTexturePersp((*p)++, G_TP_PERSP);
	gDPSetTextureFilter((*p)++, G_TF_BILERP);
	gDPSetTextureConvert((*p)++, G_TC_FILT);
	gDPSetCombineKey((*p)++, G_CK_NONE);
	gDPSetAlphaCompare((*p)++, G_AC_NONE);
	gDPSetColorDither((*p)++, G_CD_BAYER);
	gDPSetDepthImage((*p)++, zbuffer);
	const int m = 0; /* min clamp LOD frac for detail or sharpen tex mode 0 5  */
	const int l = 0; /* LOD frac for rects 0 8 */
	gDPSetPrimColor((*p)++, m, l, 0/*r*/, 0/*g*/, 64/*b*/, 255/*a*/);

	gDPPipeSync((*p)++);

	// rsp init
	{
		static Vp vp;

		vp.vp.vscale[0] = width*2; // s14.2, "width*2" = "width * 0.5(s14.2)" = half of width
		vp.vp.vscale[1] = height*2; // ditto
		vp.vp.vscale[2] = G_MAXZ/2;
		vp.vp.vscale[3] = 0;
		vp.vp.vtrans[0] = width*2;
		vp.vp.vtrans[1] = height*2;
		vp.vp.vtrans[2] = G_MAXZ/2;
		vp.vp.vtrans[3] = 0;

		gSPViewport((*p)++, &vp);
	}

	gSPClearGeometryMode((*p)++, G_SHADE | G_SHADING_SMOOTH | G_CULL_BOTH | G_FOG | G_LIGHTING | G_TEXTURE_GEN | G_TEXTURE_GEN_LINEAR | G_LOD | G_ZBUFFER);
	gSPSetGeometryMode((*p)++, G_CULL_BACK);
}}}

void draw_ovl_string(Gfx **p, const char *str) {{{
	for(const unsigned char *c = reinterpret_cast<const unsigned char*>(str); *c; c++)
	{
		if(*c != ' ') {
			gDPLoadTextureBlock((*p)++, _binary_fontia44_raw_start + *c * (16*16), G_IM_FMT_IA, G_IM_SIZ_8b, 16, 16, 0, G_TX_CLAMP | G_TX_NOMIRROR, G_TX_CLAMP | G_TX_NOMIRROR, 4/*mask s*/, 4/*mask t*/, G_TX_NOLOD, G_TX_NOLOD);
			{
				gSPVertex((*p)++, g_vs_rect16, 4/*numverts*/, 0/*dest SP vert first index*/);
				gSP1Triangle((*p)++, 2, 1, 0, 0);
				gSP1Triangle((*p)++, 3, 2, 0, 0);
			}
		}
		gSPMatrix((*p)++, IMTOPHYS(&g_mtx_move_right), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
	}
}}}

void draw_ovl_start(Gfx **p) {{{
	{
		gSPMatrix((*p)++, IMTOPHYS(&g_mtx_e), G_MTX_MODELVIEW|G_MTX_LOAD|G_MTX_NOPUSH);
		gSPMatrix((*p)++, IMTOPHYS(&g_mtx_e), G_MTX_PROJECTION|G_MTX_LOAD|G_MTX_NOPUSH);
	}
	{
		static Mtx smt, sms;
		guTranslate(&smt, -1.00, 1.00, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		guScale(&sms, 2/320.00, -2/240.00, 1.00);
		data_cache_hit_writeback(&sms, sizeof sms);
		gSPMatrix((*p)++, IMTOPHYS(&sms), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
	}
	gSPTexture((*p)++, 0x8000/*s scale .16=0.5*/, 0x8000/*t scale .16=0.5*/, 0/*max mipmap level*/, G_TX_RENDERTILE, G_ON);
	gDPSetCombineMode((*p)++, G_CC_MODULATEIA_PRIM, G_CC_MODULATEIA_PRIM);
	gSPClearGeometryMode((*p)++, G_SHADE | G_LIGHTING | G_SHADING_SMOOTH | G_ZBUFFER);

	gDPSetRenderMode((*p)++, G_RM_XLU_SURF, G_RM_XLU_SURF2); // no-aa, no-zbuf, no-mipmap, translucent-surface
	gDPSetTextureFilter((*p)++, G_TF_POINT);
}}}

void draw_common_ovl(Gfx **p) {{{
	{
		static Mtx smt, smr;
		guTranslate(&smt, 10.00*32, 16.00*0, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		guRotate(&smr, 90.0, 0.00, 0.00, 1.00);
		data_cache_hit_writeback(&smr, sizeof smr);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		gSPMatrix((*p)++, IMTOPHYS(&smr), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);

		{
			static Mtx smt;
			guTranslate(&smt, 10.00*0, 16.00*0, 0.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 0/*r*/, 255/*g*/, 0/*b*/, 255/*a*/);
			draw_ovl_string(p, "HIGH");
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}

		{
			static Mtx smt;
			guTranslate(&smt, 10.00*5, 16.00*0, 0.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 0/*b*/, 255/*a*/);
			{
				char buf[32];
				sprintf(buf, "%06d", g_highscore);
				draw_ovl_string(p, buf);
			}
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}

		if((g_curscene == &g_scene_title) || (g_count / 15 % 2 == 0)) {
			static Mtx smt;
			guTranslate(&smt, 10.00*0, 16.00*1, 0.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 0/*g*/, 0/*b*/, 255/*a*/);
			draw_ovl_string(p, "1UP");
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}

		{
			static Mtx smt;
			guTranslate(&smt, 10.00*4, 16.00*1, 0.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, 255/*a*/);
			{
				char buf[32];
				sprintf(buf, "%06d", g_score);
				draw_ovl_string(p, buf);
			}
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}

		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

#ifdef DEBUG
	if(debugon) {
		gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, 255/*a*/);
		{
			static Mtx smt;
			guTranslate(&smt, 0.00, 16.00*12, 0.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			{
				char buf[32];
				//extern char end[];
				sprintf(buf, "U=%dK G=%d/%d M=%d"
						,((char*)sbrk(0) - (char*)0x80000000) / 1024
						,debuggfxcount
						,debuggfxmax
						,g_curmtx - g_mtxbuf
					   );
				draw_ovl_string(p, buf);
			}
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}
		{
			static Mtx smt;
			guTranslate(&smt, 0.00, 16.00*13, 0.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			{
				char buf[32];
				int bu = 0;
				for(auto &bullet : g_scene_game.bullets) {
					bu += bullet.enabled;
				}
				sprintf(buf, "DX=%d DY=%d B=%d/%d DS=%d"
						,pdebugx
						,pdebugy
						,g_scene_game.mybullets
						,bu
						,SND1::debugsndcount
					   );
				draw_ovl_string(p, buf);
			}
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}
		for(int i = 0; i < 8; i++) {
			Mtx *smt = mtx();
			guTranslate(smt, 10.00 * i, 16.00*14, 0.00);
			gSPMatrix((*p)++, IMTOPHYS(smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			{
				char buf[32];
				sprintf(buf, "%d", Voice_Stopped(8+i));
				draw_ovl_string(p, buf);
			}
			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}
	}
#endif
}}}

void draw_common_ingame_ovl(Gfx **p) {{{
	gSPTexture((*p)++, 0x8000/*s scale .16=0.5*/, 0x8000/*t scale .16=0.5*/, 0/*max mipmap level*/, G_TX_RENDERTILE, G_OFF);
	gDPSetCombineMode((*p)++, G_CC_PRIMITIVE, G_CC_PRIMITIVE);

	{
		static Mtx smt, sms, smtup;

		guTranslate(&smt, 2.00, 2.00 + 3.0*39, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		guScale(&sms, 14.00, 2.00, 1.00);
		data_cache_hit_writeback(&sms, sizeof sms);
		gSPMatrix((*p)++, IMTOPHYS(&sms), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);

		guTranslate(&smtup, 0.00, -1.50, 0.00);
		data_cache_hit_writeback(&smtup, sizeof smtup);

		for(int i = 0; i < g_energy / GameScene::GE_FULL * 40; i++) {
			switch(i) {
			case 0:
				gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 0/*g*/, 0/*b*/, 255/*a*/);
				break;
			case GameScene::GE_REDLINE / GameScene::GE_FULL * 40:
				gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 0/*b*/, 255/*a*/);
				break;
			case GameScene::GE_WARNLINE / GameScene::GE_FULL * 40:
				gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 64/*r*/, 64/*g*/, 255/*b*/, 255/*a*/);
				break;
			default:
				break;
			}

			gSPVertex((*p)++, g_vs_rect1, 4/*numverts*/, 0/*dest SP vert first index*/); // note: matrix multiplied at end of loop, vertex cache must be reloaded
			gSP1Triangle((*p)++, 2, 1, 0, 0);
			gSP1Triangle((*p)++, 3, 2, 0, 0);
			gSPMatrix((*p)++, IMTOPHYS(&smtup), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		}

		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	gSPTexture((*p)++, 0x8000/*s scale .16=0.5*/, 0x8000/*t scale .16=0.5*/, 0/*max mipmap level*/, G_TX_RENDERTILE, G_ON);
	gDPSetCombineMode((*p)++, G_CC_MODULATEIA_PRIM, G_CC_MODULATEIA_PRIM);

	gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, 255/*a*/);

	{
		static Mtx smt, smr;
		guTranslate(&smt, 2.00, 10.00*18 + 5.0, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		guRotate(&smr, 90.0, 0, 0, -1);
		data_cache_hit_writeback(&smr, sizeof smr);
		gSPMatrix((*p)++, IMTOPHYS(&smr), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		draw_ovl_string(p, "ENERGY");
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	{
		static Mtx smt;
		guTranslate(&smt, 16.00*1 + 2.0, 16.00*0 + 2.0, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		gDPLoadTextureBlock((*p)++, _binary_player_data_start, G_IM_FMT_RGBA, G_IM_SIZ_32b, 32, 32, 0, G_TX_CLAMP | G_TX_NOMIRROR, G_TX_CLAMP | G_TX_NOMIRROR, 5/*mask s*/, 5/*mask t*/, G_TX_NOLOD, G_TX_NOLOD);

		for(int i = 0; i < g_lives; i++)
		{
			gSPVertex((*p)++, g_vs_rect32, 4/*numverts*/, 0/*dest SP vert first index*/); // note: matrix multiplied at end of loop, vertex cache must be reloaded
			gSP1Triangle((*p)++, 2, 1, 0, 0);
			gSP1Triangle((*p)++, 3, 2, 0, 0);
			gSPMatrix((*p)++, IMTOPHYS(&g_mtx_move_down), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		}
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
}}}

void mute_player() {{{
	g_slaser.stop();
	g_smyguard.stop();
	g_salert.stop();
	g_smyref1.stop();
	g_smyref2.stop();
}}}

void mute_all() {{{
	mute_player();
	g_senedown.stop();
}}}

void save_reset() {{{
	memset(&g_save, 0, sizeof(g_save));
	g_save.s.highscore =
#ifdef DEBUG
		debugon ? 150 :
#endif
		15000;
}}}

void save_reload() {{{
	g_highscore = g_save.s.highscore;
}}}

void save_store() {{{
	g_save.s.highscore = g_highscore;
}}}

void GameScene::draw(Gfx **p) {{{
	gDPSetRenderMode((*p)++, G_RM_ZB_OPA_SURF, G_RM_ZB_OPA_SURF2); // noaa, use-zbuf, no-mipmap, opaque-surface

	{
		u16 perspNorm;
		static Mtx proj, view;
		static Lights1 lights = gdSPDefLights1(0,0,0, 255,255,255,0,0,127);
		guPerspective(&proj, &perspNorm, 33.00/*fovy*/, 1.33/*aspect*/, 10.00/*near*/, 10000.00/*far*/, 1.0/*scale*/);
		guLookAt(&view, /*Eye*/650.00+eyex,-548.00+eyey,2099.00+eyez, /*At*/650.00+eyex,-547.50+eyey,2098.13+eyez, /*Up*/0.00,0.87,0.50);
		data_cache_hit_writeback(&proj, sizeof proj);
		gSPMatrix((*p)++, IMTOPHYS(&proj), G_MTX_PROJECTION|G_MTX_LOAD|G_MTX_NOPUSH);
		data_cache_hit_writeback(&view, sizeof view);
		gSPMatrix((*p)++, IMTOPHYS(&view), G_MTX_PROJECTION|G_MTX_MUL|G_MTX_NOPUSH);
		gSPPerspNormalize((*p)++, perspNorm);
		gSPSetLights1((*p)++, lights);
		gSPSetGeometryMode((*p)++, G_LIGHTING);
	}
	{
		gSPMatrix((*p)++, IMTOPHYS(&g_mtx_e), G_MTX_MODELVIEW|G_MTX_LOAD|G_MTX_NOPUSH);
	}

	{
		draw_object_Plane(p);
	}
	{
		draw_object_wall(p);
	}
	{
		draw_object_entry(p);
	}

	{
		{
			static Mtx smt;
			guTranslate(&smt, 0.00, 0.00, entryblkz);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		}
		draw_object_entryblk(p);
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	for(auto &block : blocks) {
		Mtx *pmt = mtx();
		guTranslate(pmt, block.pt.x, block.pt.y, 0.00);
		gSPMatrix((*p)++, IMTOPHYS(pmt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		draw_object_block(p);
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	for(auto &tank : tanks) {
		if(!tank.enabled) {
			continue;
		}

		{
			Mtx *pmt = mtx(), *pmr = mtx();
			guTranslate(pmt, tank.pt.x, tank.pt.y, 0.00);
			gSPMatrix((*p)++, IMTOPHYS(pmt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			guRotate(pmr, tank.rot, 0.00, 0.00, 1.00);
			gSPMatrix((*p)++, IMTOPHYS(pmr), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		}
		if(tank.boomstate != BS_NONE && tank.boomstate != BS_DONE) {
			tank.boomviewcount++;
			{
				Mtx *pms = mtx();
				float s;
				if(tank.boomstate == BS_EXPLODING) {
					//s = 0.5f + tank.boomstatecount / 120.0;
					s = tank.boomstatecount / 60.0;
					s = 1 - s;
					s = s*s;
					s = 1 - s;
					s = 0.5f + s/2;
					s *= 1.33f;
				} else /*if(tank.boomstate == BS_IMPLODING)*/ {
					//s = 1.0f - (tank.boomstatecount / 240.0);
					s = tank.boomstatecount / 120.0;
					s = s*s;
					s = 1 - s;
					s = 0.5f + s/2;
					s *= 1.33f;
				}
				guScale(pms, s, s, s);
				gSPMatrix((*p)++, IMTOPHYS(pms), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
				draw_object_boom(p, tank.boomviewcount * 128 % 2048);
				gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
			}
		}
		switch(tank.type) {
			case TT_PLAYER:
				if(tank.boomstate <= BS_EXPLODING) {
					const struct Color *col = &g_rainbowcolors[tank.count[0] % 8];
					gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, col->r/*r*/, col->g/*g*/, col->b/*b*/, 128/*a*/);
					draw_object_mytank(p, tank.count[1] % (32<<6), tank.headrot - tank.rot);
				}
				break;
			case TT_WEAK:
				if(tank.boomstate <= BS_EXPLODING) {
					draw_object_enetank(p, tank.count[1] % (32<<6));
				} else {
					draw_object_enetank_broken1(p);
				}
				break;
			case TT_STRONG:
			case TT_CHASE:
			case TT_CHASE_FAST:
			case TT_SUPER:
				break;
		}
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	gDPSetRenderMode((*p)++, G_RM_ZB_XLU_SURF, G_RM_ZB_XLU_SURF2); // no-aa, zbuf, no-mipmap, translucent-surface
	{
		for(auto &bullet : bullets)
		{
			if(!bullet.enabled) {
				continue;
			}

			{
				Mtx *pmt = mtx(), *pmr = mtx();
				guTranslate(pmt, bullet.pt.x, bullet.pt.y, 30.00);
				gSPMatrix((*p)++, IMTOPHYS(pmt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
				guRotate(pmr, bullet.rot, 0.00, 0.00, 1.00);
				gSPMatrix((*p)++, IMTOPHYS(pmr), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
			}

			switch(bullet.type) {
			case BT_PLAYER:
				bullet.count++;
				{
					const struct Color *col = &g_rainbowcolors[bullet.count % 8];
					gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, col->r/*r*/, col->g/*g*/, col->b/*b*/, 128/*a*/);
					draw_object_mylaser(p);
				}
				break;
			case BT_WEAK:
				{
					draw_object_enelaser(p);
				}
			case BT_STRONG:
			case BT_SUPERSTRONG:
			case BT_CHASE:
				break;
			case BT_PLAYERHIT:
				bullet.count++;
				{
					const struct Color *col = &g_rainbowcolors[bullet.count % 8];
					gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, col->r/*r*/, col->g/*g*/, col->b/*b*/, 128/*a*/);
					draw_object_hitsphere(p);
				}
				break;
			case BT_ENEMYHIT:
				break;
			}

			gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
		}
	}

	if(guarding) {
		Tank &me = tanks[0];
		static Mtx smt, smr;
		const struct Color *col = &g_rainbowcolors[static_cast<int>(std::max(0.0f, std::min(6.0f, ((g_energy - GE_REDLINE) / (GE_FULL - GE_REDLINE)) * 7))) + me.count[0] % 2];
		guTranslate(&smt, me.pt.x, me.pt.y, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		guRotate(&smr, me.count[0] * 147.0f, 0, 0, 1);
		data_cache_hit_writeback(&smr, sizeof smr);
		gSPMatrix((*p)++, IMTOPHYS(&smr), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, col->r/*r*/, col->g/*g*/, col->b/*b*/, 128/*a*/);
		gSPClearGeometryMode((*p)++, G_CULL_BACK);
		draw_object_mybarrier(p);
		gSPSetGeometryMode((*p)++, G_CULL_BACK);
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	draw_ovl_start(p);
	draw_common_ingame_ovl(p);
	draw_common_ovl(p);
}}}

void draw(void* cfb) {{{
	Gfx *p = (Gfx*)glist;
	const int scrWidth = 320;
	const int scrHeight = 240;

	mtx_rewind();

	// use only segment 0... 4:ign,4:seg,24:off -> 0x80xxxxxx -> always seg0. ok.
	gSPSegment(p++, 0, 0x00000000);

	draw_setup(&p, scrWidth, scrHeight);

	/// {{{ clear buffers
	// clear z buffer if use Z
#if 1
	gDPSetColorImage(p++, G_IM_FMT_RGBA, G_IM_SIZ_16b/*zbuf is 16b!*/, scrWidth, IMTOPHYS(zbuffer));
	gDPPipeSync(p++);
	gDPSetRenderMode(p++, G_RM_NOOP, G_RM_NOOP2);
	gDPSetCycleType(p++, G_CYC_FILL);
	gDPSetFillColor(p++, GPACK_ZDZ(G_MAXFBZ, 0) << 16 | GPACK_ZDZ(G_MAXFBZ, 0));
	gDPFillRectangle(p++, 0, 0, scrWidth-1, scrHeight-1);
	gDPPipeSync(p++);
#endif

	// clear frame buffer
	gDPSetRenderMode(p++, G_RM_NOOP, G_RM_NOOP2);
	gDPSetCycleType(p++, G_CYC_FILL);
	gDPSetColorImage(p++, G_IM_FMT_RGBA, G_IM_SIZ_16b/*set same as VI mode*/, scrWidth, IMTOPHYS(cfb));
	{
		/*
		static char kolor = 0;
		kolor = (kolor + 1) & 0x3F;
		unsigned char kkolor = kolor << 1;
		gDPSetFillColor(p++, GPACK_RGBA5551(0,kkolor + 64,64,1) << 16 | GPACK_RGBA5551(0,kkolor + 64,64,1)); // for 16b
		//gDPSetFillColor(p++, GPACK_RGBA8888(0,kolor,64,1)); // for 32b
		*/
		u16 kolor = GPACK_RGBA5551(0,0,0,1);
		gDPSetFillColor(p++, (kolor << 16) | kolor); // for 16b
	}
	gDPFillRectangle(p++, 0, 0, scrWidth-1, scrHeight-1);
	gDPPipeSync(p++);
	gDPSetCycleType(p++, G_CYC_1CYCLE);
	/// }}}

	// do draw!
	gSPSetGeometryMode(p++, G_ZBUFFER);
	gSPClipRatio(p++, FRUSTRATIO_2); // set scissor-clip ratio. 1~6

	g_curscene->draw(&p);

	// end
	gDPFullSync(p++); // only for last
	gSPEndDisplayList(p++);

#ifdef DEBUG
	// TODO: assert?
	if(countof(glist) < (unsigned int)(p - glist)) {
		_exit(1);
	}

	// TODO: assert?
	if(countof(g_mtxbuf) < (unsigned int)(g_curmtx - g_mtxbuf)) {
		_exit(1);
	}

	debuggfxcount = p - glist;
	debuggfxmax = std::max(debuggfxmax, debuggfxcount);
	init_gfx_task(&gfxtask, dram_stack, output_buf, glist, 0xC0000000 + (unsigned int)(p - glist));
#endif

	data_cache_hit_writeback(glist, (char*)p - (char*)glist);
	mtx_writeback();
	runtask(&gfxtask);
	waittask();
}}}

void game_start() {{{
	g_score = 0;
	g_lives =
#ifdef DEBUG
		debugon ? 1 :
#endif
		3;
	g_battling = 1;
	g_highscored = false;

	srand(get_ticks());

	scene_next(&g_scene_prebattling);
}}}

void GameScene::init() {{{
	gametimer = -60;
	fintimer = -1;
	reftimer = -1;

	g_lives--;
	mybullets = 0;
	entryblkz = -200;

	guarding = false;
	shooting = false;
	myspeed = 0;

	Tank &me = tanks[0];
	me.enable(TT_PLAYER, Point(650, 0), 90.0); // y+
	me.count[0] = 0;
	me.count[1] = 0;

	for(size_t i = 1; i < tanks.size(); i++) {
		tanks[i].enabled = false;
	}
	for(auto &bullet : bullets) {
		bullet.enabled = false;
	}

	tanks[1].enable(TT_WEAK, Point(   0,  350),   0.0); // x+
	tanks[2].enable(TT_WEAK, Point(   0,  750),   0.0); // x+
	tanks[3].enable(TT_WEAK, Point(   0, 1150),   0.0); // x+
	tanks[4].enable(TT_WEAK, Point( 350, 1500), 270.0); // y-
	tanks[5].enable(TT_WEAK, Point( 950, 1500), 270.0); // y-
	tanks[6].enable(TT_WEAK, Point(1300, 1150), 180.0); // x-
	tanks[7].enable(TT_WEAK, Point(1300,  750), 180.0); // x-
	tanks[8].enable(TT_WEAK, Point(1300,  350), 180.0); // x-

	blocks.clear();

	switch((g_battling - 1) % 5) {
	case 0: // battling 1
		blocks.push_back(Block(Point( 350,  250)));
		blocks.push_back(Block(Point( 950,  300)));
		blocks.push_back(Block(Point( 200,  450)));
		blocks.push_back(Block(Point(1050,  950)));
		blocks.push_back(Block(Point( 300, 1200)));
		break;
	case 1: // battling 2
		blocks.push_back(Block(Point( 550,  250)));
		blocks.push_back(Block(Point( 400,  300)));
		blocks.push_back(Block(Point( 400,  400)));
		blocks.push_back(Block(Point( 500,  450)));
		blocks.push_back(Block(Point( 200,  550)));
		blocks.push_back(Block(Point( 250,  900)));
		blocks.push_back(Block(Point( 250, 1000)));
		blocks.push_back(Block(Point( 650, 1000)));
		blocks.push_back(Block(Point( 750, 1000)));
		blocks.push_back(Block(Point( 350, 1050)));
		blocks.push_back(Block(Point( 250, 1100)));
		blocks.push_back(Block(Point( 650, 1200)));
		blocks.push_back(Block(Point( 600, 1300)));
		blocks.push_back(Block(Point( 700, 1300)));
		blocks.push_back(Block(Point( 650, 1400)));
		blocks.push_back(Block(Point( 750, 1450)));
		break;
	case 2: // battling 3
		blocks.push_back(Block(Point( 650,  600)));
		blocks.push_back(Block(Point( 750,  700)));
		blocks.push_back(Block(Point( 550,  800)));
		blocks.push_back(Block(Point( 650,  800)));
		break;
	case 3: // battling 4
		blocks.push_back(Block(Point( 550,  450)));
		blocks.push_back(Block(Point( 650,  500)));
		blocks.push_back(Block(Point( 500,  550)));
		blocks.push_back(Block(Point( 750,  550)));
		blocks.push_back(Block(Point( 600,  600)));
		blocks.push_back(Block(Point( 850,  600)));
		blocks.push_back(Block(Point( 450,  650)));
		blocks.push_back(Block(Point( 700,  650)));
		blocks.push_back(Block(Point( 950,  650)));
		blocks.push_back(Block(Point( 550,  700)));
		blocks.push_back(Block(Point( 800,  700)));
		blocks.push_back(Block(Point( 400,  750)));
		blocks.push_back(Block(Point( 650,  750)));
		blocks.push_back(Block(Point( 900,  750)));
		blocks.push_back(Block(Point( 500,  800)));
		blocks.push_back(Block(Point( 750,  800)));
		blocks.push_back(Block(Point( 350,  850)));
		blocks.push_back(Block(Point( 600,  850)));
		blocks.push_back(Block(Point( 850,  850)));
		blocks.push_back(Block(Point( 450,  900)));
		blocks.push_back(Block(Point( 700,  900)));
		blocks.push_back(Block(Point( 550,  950)));
		blocks.push_back(Block(Point( 800,  950)));
		blocks.push_back(Block(Point( 650, 1000)));
		blocks.push_back(Block(Point( 750, 1050)));
		break;
	case 4: // battling 5
		blocks.push_back(Block(Point( 850,  250)));
		blocks.push_back(Block(Point( 450,  300)));
		blocks.push_back(Block(Point( 650,  650)));
		blocks.push_back(Block(Point( 350,  750)));
		blocks.push_back(Block(Point( 600, 1150)));
		break;
	}

	mute_all();
	g_sgetready.play();
}}}

bool GameScene::canmove(Tank *me, Point target, bool *pslow) {{{
	static const Rect worldrc(50, 50, 1250, 1450);
	static const float thresh = 90;

	if(!worldrc.inrect(target)) {
		return false;
	}

	for(auto &block : blocks) {
		if(block.pt.incircle(target, thresh)) {
			return false;
		}
	}

	for(auto &tank : tanks) {
		if(!tank.enabled) {
			continue;
		}
		if(&tank == me) {
			continue;
		}
		if(tank.pt.incircle(target, thresh)) {
			if(tank.boomstate != BS_DONE) {
				return false;
			} else { // ==BS_DONE
				if(tank.type != TT_PLAYER && pslow) { // TT_PLAYER does not slow tanks
					*pslow = true;
					pslow = nullptr; // no more slowdown
				}
			}
		}
	}

	return true;
}}}

bool GameScene::can_move_then_move(Tank *me, float rotr, float len, bool *pslow, float olen, float slowcoeff) {{{
	bool can, slow = false;

	Point pt = me->pt;
	pt.add_angular(rotr, len);

	me->count[1] += olen * 16;
	can = canmove(me, pt, &slow);
	if(pslow) {
		*pslow = slow;
	}
	if(can) {
		// recalculate because len may changed.
		// TODO: fix this ugly non-DRY code
		Point pt = me->pt;
		pt.add_angular(rotr, len / (slow ? slowcoeff : 1));

		// TODO may be not required...
		me->pt.x = std::max(50.f, std::min(1250.f, pt.x));
		me->pt.y = std::max(50.f, std::min(1450.f, pt.y));
	}
	return can;
}}}

void GameScene::input() {{{
	struct controller_data pkeys;
	struct SI_condat *p0;
	int controllers;

	controller_scan();
	controllers = get_controllers_present();

	if(!(controllers & CONTROLLER_1_INSERTED)) {
		return;
	}

	if(fintimer == 0) {
		return;
	}

	pkeys = get_keys_pressed();
	p0 = &pkeys.c[0];

#ifdef DEBUG
	pdebugx = p0->x;
	pdebugy = p0->y;
#endif

	Tank &me = tanks[0];

	if(0 <= gametimer && me.boomstate == BS_NONE) {
		shooting = false;
		if(p0->A) {
			if(GED_SHOOTING < g_energy) {
				shooting = true;

				if(mybullets < 8) {
					for(auto &bullet : bullets) {
						if(bullet.enabled) {
							continue;
						}
						bullet.enable(BT_PLAYER, tanks[0].pt, me.headrot, 60);
						bullet.count = mybullet_color;
						mybullet_color += 7;
						mybullets++;

						g_slaser.play();

						break;
					}
				} else { // FIXME umm... g_slaser.stop()s looks bad.
					g_slaser.stop();
				}
			} else {
				g_slaser.stop();
			}
		} else {
			g_slaser.stop();
		}

		guarding = p0->B && GE_REDLINE < g_energy;
		if(guarding) {
			if(GE_ALERT < g_energy) {
				g_smyguard.play();
			}
		} else {
			g_smyguard.stop();
		}

		if(p0->x != 0 || p0->y != 0)
		{
			const static Point zero(0, 0);
			const float thresh = 10;
			float rotr = atan2(p0->y, p0->x);
			float rot = degnormalize(rad2deg(rotr));
			float len = std::min(126.0f, zero.distance(Point(p0->x, p0->y))) / 126.0f * 5;
			float movingdeg = smallerdeg(me.headrot, rot);
			float olen;

			moving = true;

			if(myspeed == 0) {
				mymomentumrot = rot;
			} else if(mymomentumrot != rot) {
				float friction = smallerdeg(mymomentumrot, rot) / 180.0f;
				float smoothness = 1.0 - friction;
				myspeed *= smoothness * smoothness;
				mymomentumrot = rot;
			}

			olen = len;
			len *= myspeed;

			if(movingdeg < thresh) {
				bool slow = false;

				// slightly rotate: rotate head and tank, then forward.
				me.rot = me.headrot = rot;

				// TODO: natural rotating tank, not only head.
				can_move_then_move(&me, rotr, len, &slow, olen, 2.0f);
				if(slow && 0.5 < myspeed) {
					myspeed /= 1.5;
				}
			} else {
				// more rotate: only rotating head
				me.headrot = degnormalize(me.headrot + thresh * nearside(me.headrot, rot));
			}
		} else {
			moving = false;
		}
	}
}}}

void GameScene::bullethit(Bullet *bullet, bool quick) {{{
	if(bullet->type == BT_PLAYER) {
		mybullets--;
	}

	if(quick) {
		bullet->enabled = false;
		return;
	}

	if(bullet->type == BT_PLAYER) {
		bullet->type = BT_PLAYERHIT;
	} else {
		bullet->type = BT_ENEMYHIT;
	}
}}}

bool GameScene::isfin() {{{
	for(auto &tank : tanks) {
		if(!tank.enabled) {
			continue;
		}
		if(tank.boomstate != BS_NONE) {
			continue;
		}
		switch(tank.type) {
		case TT_PLAYER:
			continue;
		case TT_WEAK:
		case TT_STRONG:
		case TT_CHASE:
		case TT_CHASE_FAST:
		case TT_SUPER:
			return false;
		}
	}

	return true;
}}}

void GameScene::fincheck() {{{
	if(fintimer != -1) {
		return;
	}

	if(!isfin()) {
		return;
	}

	fintimer = 190;
}}}

void GameScene::boom(Tank *tank, int chain) {{{
	if(tank->type == TT_PLAYER && guarding) {
		if(0 < chain) {
			// chainboom
			if(GE_WARNLINE < g_energy) {
				g_energy -= GED_BOOM;
				return;
			}
		} else {
			// boom by bullet
			if(GE_REDLINE < g_energy) {
				g_energy -= GED_LASER;
				reftimer = 3;
				g_smyref2.stop();
				g_smyref1.play();
				return;
			}
		}
	}

	tank->boom();
	if(tank->type == TT_PLAYER) { // enemy to me
		mute_player();
		guarding = false;

		Player_Stop();
		Player_Start(g_module);
		Player_SetPosition(0); // reset
		Player_SetPosition(1);

		if(fintimer == -1) {
			fintimer = 90;
		}
	} else { // me to enemy
		g_senedown.stop();
		g_senedown.play();
		if(0 < chain) {
			tank->chain = chain + 1;
		}
		fincheck();
	}
}}}

void GameScene::hittest(Bullet *bullet) {{{
	if(bullet->pt.x < 0 || bullet->pt.y < 0 || 1300 < bullet->pt.x || 1500 < bullet->pt.y) {
		bullet->pt.x = std::max(0.f, std::min(1300.f, bullet->pt.x));
		bullet->pt.y = std::max(0.f, std::min(1500.f, bullet->pt.y));
		bullethit(bullet);
		return;
	}

	for(auto &block : blocks) {
		if(block.pt.incircle(bullet->pt, 50)) { // TODO should be rect-hittest
			bullethit(bullet);
			return;
		}
	}

	for(auto &tank : tanks) {
		if(!tank.enabled) {
			continue;
		}

		bool hit = tank.boomstate == BS_NONE;
		if(bullet->type == BT_PLAYER) {
			hit &= tank.type != TT_PLAYER;
		} else {
			hit &= tank.type == TT_PLAYER;
		}
		hit &= tank.pt.incircle(bullet->pt, (tank.type == TT_PLAYER && guarding) ? 100 : 50);
		if(hit) {
			bullethit(bullet, true);
			boom(&tank);
			return;
		}
	}
}}}

void GameScene::think_weak(Tank *tank) {{{
	if(tank->count[0] == 0) {
		float len = 3;
		Point pt = tank->pt;
		pt.add_angular(deg2rad(tank->rot), len);

		if(!canmove(tank, pt, nullptr) && false/*TODO DEBUG*/) {
			// if current rot is blocked...
			// TODO search where is open
			//      turn nearest angle
			//      limit 45deg
		} else {
			// not blocked, turn head to player
			Point dp = tanks[0].pt;
			dp -= tank->pt;
			float drot = degnormalize(rad2deg(atan2(dp.y, dp.x)));
			float nearrot = nearside(tank->rot, drot) * std::min(45.f, smallerdeg(tank->rot, drot));
			tank->rot = degnormalize(tank->rot + nearrot);
		}
	}

	if((std::max(0, 50 - g_battling * 10)) < gametimer && tank->count[0] % 5 == 0 && rand() < RAND_MAX / std::max(2, (10 - g_battling) * 7)) {
		for(auto &bullet : bullets) {
			if(bullet.enabled) {
				continue;
			}
			bullet.enable(BT_WEAK, tank->pt, tank->rot, 30);
			break;
		}
	}

	can_move_then_move(tank, deg2rad(tank->rot), 3, nullptr, 3, 4.0f);
	tank->count[0] = (tank->count[0] + 1) % 50;
}}}

void add_score(int addition) {{{
	if((g_score < SC_FIRSTEXT && SC_FIRSTEXT <= g_score + addition)
		|| (SC_FIRSTEXT < g_score && ((g_score + addition - SC_FIRSTEXT) % SC_EVERYEXT < (g_score - SC_FIRSTEXT) % SC_EVERYEXT))) {
		g_lives++;

		g_sextend.play();
	}

	g_score += addition;

	if(g_highscore < g_score) {
		g_highscore = g_score;
		if(!g_highscored) {
			g_highscored = true;

			g_shighscore.play();
		}
	}
}}}

void GameScene::think() {{{
	Tank &me = tanks[0];

	g_senedown.refresh();

	if(fintimer != -1) {
		if(fintimer == 0) {
			// add score of tanks that is (ex|im)ploding before fin (at this time)
			for(auto &tank : tanks) {
				if(!tank.enabled || tank.type == TT_PLAYER) {
					continue;
				}
				if(tank.boomstate == BS_EXPLODING || tank.boomstate == BS_IMPLODING) {
					add_score(100 + 400 * (tank.chain - 1));
				}
			}

			if(me.boomstate == BS_NONE) {
				// battling clear, advance.
				g_lives++; // restore
				g_battling++; // advance
				scene_next(&g_scene_prebattling);
			} else {
				// dead.
				if(g_lives == 0) {
					// game over
					scene_next(&g_scene_gameover);
				} else {
					// battle again
					if(isfin()) { // if dead but all enemies destroyed,
						// no lives restore.
						g_battling++; // advance.
					}
					scene_next(&g_scene_prebattling);
				}
			}
			return;
		}

		fintimer--;
	}

	gametimer++;

	if(gametimer < 0) {
		if(gametimer < -30) {
			const static Point delta[9] = {
				Point(0, 3),
				Point(3, 0), Point(3, 0), Point(3, 0),
				Point(0, -3), Point(0, -3),
				Point(-3, 0), Point(-3, 0), Point(-3, 0),
			};
			for(int i = 0; i < 9; i++) {
				tanks[i].pt += delta[i];
				tanks[i].count[1] += 3 * 16;
			}
		}
		if(-40 < gametimer && entryblkz < 0) {
			entryblkz = std::min(0, entryblkz - std::min(-1, entryblkz / 10));
		}

		tanks[0].count[0]++; // rainbowrify

		return;
	}

	if(me.boomstate == BS_NONE) {
		if(guarding) {
			g_energy = g_energy - GED_GUARDING * (moving ? GEM_MOVING : 1.0f);
			if(g_energy < 0.0f) {
				g_energy -= GE_PENALTY;
			}
		}
		if(shooting) {
			g_energy = g_energy - GED_SHOOTING * (moving ? GEM_MOVING : 1.0f);
			if(g_energy < 0.0f) {
				g_energy -= GE_PENALTY;
			}
		}
		if(!guarding && !shooting && g_energy < GE_FULL) {
			g_energy = std::min(GE_FULL, g_energy + GED_IDLEREGEN / (moving ? GEM_MOVING : 1.0f));
		}
		if(g_energy < GE_ALERT) {
			g_smyguard.stop();
			g_salert.play();
		} else {
			g_salert.stop();
		}

		if(moving) {
			myspeed = std::min(myspeed + 0.05f, 1.0f);
		} else {
			myspeed = std::max(0.0f, myspeed - 0.1f);
		}

		if(static_cast<int>(2 + (1.0f - myspeed) * 5) < ++engtimer) {
			engtimer = 0;
			g_sengine.play(myspeed * 6000 - 6000);
		}
	} else {
		g_energy = std::max(0.0f, g_energy - 2.0f);
	}

	if(0 < reftimer) {
		if(--reftimer == 0) {
			g_smyref1.stop();
			g_smyref2.play();
			reftimer = -1;
		}
	}

	for(auto &tank : tanks) {
		if(!tank.enabled) {
			continue;
		}

		if(tank.boomstate != BS_NONE) {
			switch(tank.boomstate) {
			case BS_EXPLODING:
				tank.boomstatecount += tank.type == TT_PLAYER ? 2 : 1;
				if(60 <= tank.boomstatecount) {
					tank.boomstate= BS_IMPLODING;
					tank.boomstatecount = 0;
				}

				// check chain-boom
				for(auto &otank : tanks) {
					if(!otank.enabled || &otank == &tank || otank.boomstate != BS_NONE) {
						continue;
					}
					if(tank.pt.incircle(otank.pt, (50 + tank.boomstatecount) * 1.33f)) {
						boom(&otank, tank.chain);
					}
				}
			case BS_IMPLODING:
				tank.boomstatecount += tank.type == TT_PLAYER ? 2 : 1;
				if(120 <= tank.boomstatecount) {
					tank.boomstate= BS_DONE;
					tank.boomstatecount = 0;
					if(tank.type != TT_PLAYER) {
						add_score(100 + 400 * (tank.chain - 1));
					}
				}
				break;
			default:
				break;
			}
			continue;
		}

		switch(tank.type) {
		case TT_PLAYER:
			tank.count[0]++;
			break;
		case TT_WEAK:
			think_weak(&tank);
			break;
		case TT_STRONG:
		case TT_CHASE:
		case TT_CHASE_FAST:
		case TT_SUPER:
			break;
		}
	}

	for(auto &bullet : bullets) {
		if(!bullet.enabled) {
			continue;
		}

		switch(bullet.type) {
		case BT_PLAYER:
		case BT_WEAK:
		case BT_STRONG:
		case BT_SUPERSTRONG:
			bullet.pt += bullet.delta;
			hittest(&bullet);
			break;
		case BT_CHASE:
			bullet.pt += bullet.delta; // TODO chase TT_PLAYER
			hittest(&bullet);
			break;
		case BT_PLAYERHIT:
			bullet.count++;
			if(5 < bullet.count) {
				bullet.enabled = false;
			}
			break;
		case BT_ENEMYHIT:
			bullet.count++;
			if(2 < bullet.count) {
				bullet.enabled = false;
			}
			break;
		}
	}
}}}

void PrebattlingScene::init() {{{
	g_energy = GameScene::GE_FULL;
	count =
#ifdef DEBUG
		debugon ? 1 :
#endif
		120;

	mute_all();
	Player_SetPosition(0); // note: this is important to stop all channels. Player_Stop does not stop!
	Player_Stop();
}}}

void PrebattlingScene::input() {{{
}}}

void PrebattlingScene::think() {{{
	count--;

	if(count == 0) {
		scene_next(&g_scene_game);
	}
}}}

void PrebattlingScene::draw(Gfx **p) {{{
	draw_ovl_start(p);
	{
		static Mtx smt;
		guTranslate(&smt, 10.00*11, 16.00*7, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, 255/*a*/);
		{
			char buf[32];
			sprintf(buf, "BATTLING %d", g_battling);
			draw_ovl_string(p, buf);
		}
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
	draw_common_ingame_ovl(p);
	draw_common_ovl(p);
}}}

void TitleScene::init() {{{
	bcount = 0;
	flipcount = 0;
	count = 0;
	reset_timer = 0;
}}}

void TitleScene::input() {{{
	struct controller_data pkeys, dkeys;
	struct SI_condat *p0, *d0;
	int controllers;

	controller_scan();
	controllers = get_controllers_present();

	if(!(controllers & CONTROLLER_1_INSERTED)) {
		return;
	}

	pkeys = get_keys_pressed();
	dkeys = get_keys_down();
	p0 = &pkeys.c[0];
	d0 = &dkeys.c[0];

	if(p0->L && p0->R && p0->start) {
		if(180 == ++reset_timer) {
			save_reset();
			if(!save_write(&g_save)) {
				// internal error!
				// TODO what will you do when save failed?
			}
			save_reload();

			// notice player
			Player_Start(g_module);
			Player_SetPosition(0); // reset
			Player_SetPosition(1);
		}
		return;
	} else {
		reset_timer = 0;
	}

	if(d0->A || d0->start) {
		game_start();
	}

#ifdef DEBUG
	if(d0->L) {
		debugon = !debugon;
	}
#endif

	if(d0->B && flipcount == 0) {
		if(++bcount == 10) {
			bcount = 0;
			flipcount = 180;
		}
	}
}}}

void TitleScene::think() {{{
	count++;

	if(0 < flipcount) {
		flipcount--;
	}
}}}

void TitleScene::draw(Gfx **p) {{{
	gDPSetRenderMode((*p)++, G_RM_ZB_OPA_SURF, G_RM_ZB_OPA_SURF2); // noaa, use-zbuf, no-mipmap, opaque-surface

	{
		u16 perspNorm;
		static Mtx proj, view;
		static Lights1 lights = gdSPDefLights1(0,0,0, 255,255,255,0,-63,109);
		guPerspective(&proj, &perspNorm, 33.00/*fovy*/, 1.33/*aspect*/, 10.00/*near*/, 10000.00/*far*/, 1.0/*scale*/);
		guLookAt(&view, /*Eye*/145.00,-230.00,400.00, /*At*/145.00,-229.50,399.13, /*Up*/-0.00,0.87,0.50);
		data_cache_hit_writeback(&proj, sizeof proj);
		gSPMatrix((*p)++, IMTOPHYS(&proj), G_MTX_PROJECTION|G_MTX_LOAD|G_MTX_NOPUSH);
		data_cache_hit_writeback(&view, sizeof view);
		gSPMatrix((*p)++, IMTOPHYS(&view), G_MTX_PROJECTION|G_MTX_MUL|G_MTX_NOPUSH);
		gSPPerspNormalize((*p)++, perspNorm);
		gSPSetLights1((*p)++, lights);
		gSPSetGeometryMode((*p)++, G_LIGHTING);
	}
	{
		gSPMatrix((*p)++, IMTOPHYS(&g_mtx_e), G_MTX_MODELVIEW|G_MTX_LOAD|G_MTX_NOPUSH);
	}

	{
		{
			static Mtx smt, smr;
			float rot;
			guTranslate(&smt, 145.00, 0.00, 10.00);
			data_cache_hit_writeback(&smt, sizeof smt);
			gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
			rot = flipcount / 180.0f;
			rot = rot * rot * rot;
			rot = rot * (360*3);
			guRotate(&smr, rot, 0.00, 1.00, 0.00);
			data_cache_hit_writeback(&smr, sizeof smr);
			gSPMatrix((*p)++, IMTOPHYS(&smr), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		}
		draw_object_title(p);
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}

	draw_ovl_start(p);
	draw_common_ovl(p);
	gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, 255/*a*/);
	{
		static Mtx smt;
		guTranslate(&smt, 10.00*26, 16.00*7, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		draw_ovl_string(p, "2015");
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
	{
		unsigned char a = static_cast<unsigned char>(abs(sin(count / 30.0f * M_PI) * 255));
		//unsigned char a = count%255;
		gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, a/*a*/);
	}
	{
		static Mtx smt;
		guTranslate(&smt, 10.00*7, 16.00*10, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		draw_ovl_string(p, "PRESS START BUTTON");
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
	gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 255/*g*/, 255/*b*/, 255/*a*/);
	{
		static Mtx smt, sms, smt2;

		guTranslate(&smt, 10.00*8, 16.00*12, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		draw_ovl_string(p, "\xA9 ");

		guTranslate(&smt2, 0.00, 16.00*0.25, 0.00);
		data_cache_hit_writeback(&smt2, sizeof smt2);
		gSPMatrix((*p)++, IMTOPHYS(&smt2), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		guScale(&sms, 1.00, 0.5, 1.0);
		data_cache_hit_writeback(&sms, sizeof sms);
		gSPMatrix((*p)++, IMTOPHYS(&sms), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		draw_ovl_string(p, "MMXIV MURACHUE");

		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
	{
		static Mtx smt, sms;

		guTranslate(&smt, 10.00*0.5, 16.00*14.25, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		guScale(&sms, 1.00, 0.5, 1.0);
		data_cache_hit_writeback(&sms, sizeof sms);
		gSPMatrix((*p)++, IMTOPHYS(&sms), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_NOPUSH);
		draw_ovl_string(p, "BUILD " __DATE__ " " __TIME__);
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
}}}

void GameoverScene::init() {{{
	count = 120;

	if(g_highscored) {
		save_store();
		if(!save_write(&g_save)) {
			// internal error!
			// TODO what will you do when save failed?
		}
	}

	mute_all();
	Player_Start(g_module);
	Player_SetPosition(0);
}}}

void GameoverScene::input() {{{
}}}

void GameoverScene::think() {{{
	count--;

	if(count == 0) {
		scene_next(&g_scene_title);
	}
}}}

void GameoverScene::draw(Gfx **p) {{{
	draw_ovl_start(p);
	{
		static Mtx smt;
		guTranslate(&smt, 10.00*12, 16.00*7, 0.00);
		data_cache_hit_writeback(&smt, sizeof smt);
		gSPMatrix((*p)++, IMTOPHYS(&smt), G_MTX_MODELVIEW | G_MTX_MUL | G_MTX_PUSH);
		gDPSetPrimColor((*p)++, 0/*min_clamp_LOD_frac*/, 0/*rect_LOD_frac*/, 255/*r*/, 0/*g*/, 0/*b*/, 255/*a*/);
		draw_ovl_string(p, "GAMEOVER");
		gSPPopMatrix((*p)++, G_MTX_MODELVIEW);
	}
	draw_common_ingame_ovl(p);
	draw_common_ovl(p);
}}}

void scene_next_do(Scene *scene) {{{
	g_curscene = scene;
	g_curscene->init();
	g_nextscene = nullptr;
}}}

int main() {{{
	sys_set_boot_cic(6102);

	/* enable interrupts (on the CPU) */
	init_interrupts();

	/* Initialize peripherals */
	display_init( RESOLUTION_320x240, DEPTH_16_BPP, 2 /*double-buffer*/, GAMMA_NONE, ANTIALIAS_RESAMPLE );
	controller_init();

	draw_init();

	//{{{ save
	save_init();
	save_pagesize(sizeof(g_save));
	save_pagecopy(2);

	if(!save_read(&g_save)) {
		// initialize
		save_reset();
	}
	save_reload();
	//}}}

	//{{{ audio
	audio_init(32000 /*Hz*/, 2 /*double-buffer*/);

	//int aulen = audio_get_buffer_length();
	//short *aubuf = (short*)malloc(aulen * 2 /*ch*/ * sizeof(short)/*bits/octet*/);

	//int aufreq = audio_get_frequency();

	//}}}

	//{{{ mod
	MikMod_RegisterDriver(&drv_n64);
	//MikMod_RegisterAllLoaders();
	MikMod_RegisterLoader(&load_xm);
	//MikMod_RegisterLoader(&ldr_it);

	md_mode = 0;
	md_mode |= DMODE_16BITS;
	md_mode |= DMODE_SOFT_MUSIC;
	md_mode |= DMODE_SOFT_SNDFX;
	md_mode |= DMODE_STEREO;

	md_mixfreq = audio_get_frequency();

	MikMod_Init("");
	MikMod_SetNumVoices(10/*music voices*/, 8/*sample voices*/);

	g_module = Player_LoadMem(_binary_Module1_xm_start, (int)_binary_Module1_xm_size, 8, 0);

	g_slaser.load(&g_module->samples[9]);
	g_senedown.load(&g_module->samples[14]);
	g_sgetready.load(&g_module->samples[17], true);
	g_smyguard.load(&g_module->samples[10]);
	g_salert.load(&g_module->samples[13]);
	g_smyref1.load(&g_module->samples[11]);
	g_smyref2.load(&g_module->samples[12]);
	g_sextend.load(&g_module->samples[19], true);
	g_shighscore.load(&g_module->samples[20], true);
	g_sengine.load(&g_module->samples[18], true);

	MikMod_EnableOutput();

	//Player_Start(g_module);
	//Player_SetPosition(2); // SetPosition effects only when playing (pf != NULL).
	//Player_Stop();
	//Player_Free(g_module);
	//}}}

	register_VI_handler(vsync_handler);

	init_gfx_task(&gfxtask, dram_stack, output_buf, glist);

	scene_next_do(&g_scene_title);

	/* Main loop */
	while(1) {
		vsync = 0;

		g_count++;

		g_curscene->input();
		g_curscene->think();

		if(audio_can_write()) {
			MikMod_Update();
		}

		display_context_t dc = display_lock();

		draw(__safe_buffer[dc - 1]);

		display_show(dc);

		if(g_nextscene) {
			scene_next_do(g_nextscene);
		}

		while(!vsync) /* wait until vsync */halt();
	}
}}}
